from .avlTree import avl_tree
import random
import math
from datetime import datetime

switch_to_algorithm_z_constant = 22
debug = 'none'


def reservoir(reservoir_size, storage_period_in_milliseconds=60000, random_number_gen=None):
    interval = storage_period_in_milliseconds / 1000
    rng = random_number_gen or random.random
    reservoir_size = max(1, (math.floor(reservoir_size) >> 0) or 1)
    total_item_count = 0
    last_deleted_index = -1
    num_to_skip = -1

    def algorithm_r():
        """"Algorithm R"
        Selects random elements from an unknown-length input.
        Has a time-complexity of: O(N)
        Number of random numbers required:
        N - n
        Where:
        n = the size of the reservoir
        N = the size of the input"""
        local_item_count = total_item_count + 1
        random_value = math.floor(rng() * local_item_count)
        to_skip = 0

        while random_value >= reservoir_size:
            to_skip += 1
            local_item_count += 1
            random_value = math.floor(rng() * local_item_count)
        nonlocal evict_next
        evict_next = random_value
        return to_skip

    def algorithm_x():
        """Selects random elements from an unknown-length input.
        Has a time-complexity of: O(N)
        Number of random numbers required:
        2 * n * ln( N / n )
        Where:
        n = the size of the reservoir
        N = the size of the input"""
        nonlocal algorithm_x_count
        nonlocal current_algorithm
        local_item_count = total_item_count
        random_value = rng()
        to_skip = 0

        if total_item_count <= switch_threshold:
            local_item_count += 1
            algorithm_x_count += 1
            quotient = algorithm_x_count / local_item_count

            while quotient > random_value:
                to_skip += 1
                local_item_count += 1
                algorithm_x_count += 1
                quotient = (quotient * algorithm_x_count) / local_item_count
            return to_skip
        else:
            current_algorithm = algorithm_z
            return current_algorithm()

    def algorithm_z():
        """"Algorithm Z"
        Selects random elements from an unknown-length input.
        Has a time-complexity of:
        O(n(1 + log (N / n)))
        Number of random numbers required:
        2 * n * ln( N / n )
        Where:
        n = the size of the reservoir
        N = the size of the input"""
        term = total_item_count - reservoir_size + 1
        nonlocal w

        while True:
            random_value = rng()
            x = total_item_count * (w - 1)
            to_skip = math.floor(x)
            subterm = ((total_item_count + 1) / term)
            subterm *= subterm
            term_skip = term + to_skip
            lhs = math.exp(math.log(((random_value * subterm) * term_skip) / (total_item_count + x)) / reservoir_size)
            rhs = (((total_item_count + x) / term_skip) * term) / total_item_count

            if lhs <= rhs:
                w = rhs / lhs
                break

            y = (((random_value * (total_item_count + 1)) / term) * (total_item_count + to_skip + 1)) / \
                (total_item_count + x)

            if reservoir_size < to_skip:
                denom = total_item_count
                numer_lim = term + to_skip
            else:
                denom = total_item_count - reservoir_size + to_skip
                numer_lim = total_item_count + 1

            for numer in range(total_item_count + to_skip, numer_lim - 1, -1):
                y = y * numer / denom
                denom -= 1

            w = math.exp(-math.log(rng()) / reservoir_size)

            if math.exp(math.log(y) / reservoir_size) <= (total_item_count + x) / total_item_count:
                break
        return to_skip

    current_algorithm = algorithm_x
    switch_threshold = switch_to_algorithm_z_constant * reservoir_size

    if debug == 'R':
        current_algorithm = algorithm_r
    elif debug == 'X':
        switch_threshold = math.inf
    elif debug == 'Z':
        current_algorithm = algorithm_z

    algorithm_x_count = 0
    w = math.exp(-math.log(rng() / reservoir_size))
    evict_next = None

    def comparator(a, b):
        if a < b:
            return -1
        if a > b:
            return 1
        return 0

    index_tree = avl_tree(lambda a, b: comparator(a['index'] if (a and 'index' in a) else 0,
                                                  b['index'] if (b and 'index' in b) else 0))
    value_tree = avl_tree(lambda a, b: comparator(a, b))
    initial_index = 0

    def remove_old_records():
        while True:
            element = index_tree['at'](0)
            if element is not None and datetime.now().timestamp() > element['time'] + interval:
                index_tree['removeAt'](0)
                nonlocal last_deleted_index
                delete_index_diff = element['index'] - last_deleted_index
                last_deleted_index = element['index']
                value_tree['remove'](element['data'])
                nonlocal total_item_count
                total_item_count -= delete_index_diff
                nonlocal algorithm_x_count
                algorithm_x_count = max(0, algorithm_x_count - delete_index_diff)
            else:
                break

    index_tree['removeOldRecords'] = remove_old_records

    def get_percentile(percent):
        index_tree['removeOldRecords']()
        index = (index_tree['size']() - 1) * percent / 100
        lower = math.floor(index)
        fraction_part = index - lower
        percentile = value_tree['at'](lower) or 0
        if fraction_part > 0:
            percentile += fraction_part * ((value_tree['at'](lower + 1) or 0) - (value_tree['at'](lower) or 0))
        return float(percentile)

    index_tree['getPercentile'] = get_percentile

    def push_some(*args):
        nonlocal initial_index
        nonlocal index_tree
        len = min(index_tree['size'](), reservoir_size)
        for arg in args:
            index_tree['removeOldRecords']()
            value = {'index': initial_index, 'time': datetime.now().timestamp(), 'data': arg}
            index_tree = add_sample(index_tree, value)
            initial_index += 1
        return len

    index_tree['pushSome'] = push_some

    def from_plain_object(*args):
        nonlocal index_tree
        nonlocal initial_index
        len = min(index_tree['size'](), reservoir_size)
        for arg in args:
            value = {'index': arg['index'], 'time': arg['time'], 'data': arg['data']}
            add_sample(index_tree, value)
            initial_index += 1
        return len

    index_tree['fromPlainObject'] = from_plain_object

    def add_sample(tree: avl_tree, sample):
        nonlocal total_item_count
        if index_tree['size']() < reservoir_size:
            tree['insert'](sample)
            value_tree['insert'](sample['data'])
        else:
            nonlocal num_to_skip
            if num_to_skip < 0:
                num_to_skip = current_algorithm()
            if num_to_skip == 0:
                replace_random_sample(sample, tree)
            num_to_skip -= 1
        total_item_count += 1
        return tree

    def replace_random_sample(sample, reservoir):
        nonlocal evict_next
        if evict_next is not None:
            random_index = evict_next
            evict_next = None
        else:
            random_index = math.floor(rng() * reservoir_size)
        value = reservoir['at'](random_index)
        reservoir['removeAt'](random_index)
        value_tree['remove'](value['data'])
        value_tree['insert'](sample['data'])
        reservoir['insert'](sample)

    return index_tree



from mock import AsyncMock, patch
import pytest
from asyncio import sleep
import asyncio
from freezegun import freeze_time
from .metaApiWebsocket_client import MetaApiWebsocketClient
from .synchronizationThrottler import SynchronizationThrottler


class MockClient(MetaApiWebsocketClient):
    async def _rpc_request(self, account_id: str, request: dict, timeout_in_seconds: float = None):
        await sleep(0.1)
        pass


start_time = '2020-10-05 10:00:00.000'
throttler: SynchronizationThrottler = None
client = None


@pytest.fixture(autouse=True)
async def run_around_tests():
    with patch('lib.clients.metaApi.synchronizationThrottler.asyncio.sleep', new=lambda x: sleep(x / 20)):
        global client
        client = MockClient('token')
        client._rpc_request = AsyncMock()
        global throttler
        throttler = SynchronizationThrottler(client, 2)
        throttler.start()
        yield
        throttler.stop()


class TestSynchronizationThrottler:
    @pytest.mark.asyncio
    async def test_sync_without_queue(self):
        """Should immediately send request if free slots exist."""
        with freeze_time() as frozen_datetime:
            frozen_datetime.move_to('2020-10-10 01:00:01.000')
            await throttler.schedule_synchronize('accountId', {'requestId': 'test'})
            assert throttler._synchronizationIds == {'test': 1602291601.0}
            throttler.remove_synchronization_id('test')
            client._rpc_request.assert_called_with('accountId', {'requestId': 'test'})
            assert throttler._synchronizationIds == {}

    @pytest.mark.asyncio
    async def test_not_remove_if_different_instance_index(self):
        """Should not remove sync if different instance index."""
        with freeze_time() as frozen_datetime:
            frozen_datetime.move_to('2020-10-10 01:00:01.000')
            await throttler.schedule_synchronize('accountId', {'requestId': 'test', 'instanceIndex': 0})
            await throttler.schedule_synchronize('accountId', {'requestId': 'test1', 'instanceIndex': 1})
            assert throttler._synchronizationIds == {'test': 1602291601.0, 'test1': 1602291601.0}
            throttler.remove_synchronization_id('test')
            assert throttler._synchronizationIds == {'test1': 1602291601.0}
            client._rpc_request.assert_any_call('accountId', {'requestId': 'test', 'instanceIndex': 0})
            client._rpc_request.assert_any_call('accountId', {'requestId': 'test1', 'instanceIndex': 1})

    @pytest.mark.asyncio
    async def test_sync_with_queue(self):
        """Should wait for other sync requests to finish if slots are full."""
        await throttler.schedule_synchronize('accountId1', {'requestId': 'test1'})
        await throttler.schedule_synchronize('accountId2', {'requestId': 'test2'})
        client._rpc_request.assert_any_call('accountId1', {'requestId': 'test1'})
        client._rpc_request.assert_any_call('accountId2', {'requestId': 'test2'})
        asyncio.create_task(throttler.schedule_synchronize('accountId3', {'requestId': 'test3'}))
        await sleep(0.1)
        assert client._rpc_request.call_count == 2
        throttler.remove_synchronization_id('test1')
        await sleep(0.1)
        assert client._rpc_request.call_count == 3

    @pytest.mark.asyncio
    async def test_not_take_slots_if_same_account(self):
        """Should not take extra slots if sync ids belong to the same account."""
        asyncio.create_task(throttler.schedule_synchronize('accountId', {'requestId': 'test', 'instanceIndex': 0}))
        asyncio.create_task(throttler.schedule_synchronize('accountId', {'requestId': 'test1', 'instanceIndex': 1}))
        asyncio.create_task(throttler.schedule_synchronize('accountId2', {'requestId': 'test2'}))
        asyncio.create_task(throttler.schedule_synchronize('accountId3', {'requestId': 'test3'}))
        await sleep(0.2)
        assert client._rpc_request.call_count == 3
        client._rpc_request.assert_any_call('accountId', {'requestId': 'test', 'instanceIndex': 0})
        client._rpc_request.assert_any_call('accountId', {'requestId': 'test1', 'instanceIndex': 1})
        client._rpc_request.assert_any_call('accountId2', {'requestId': 'test2'})

    @pytest.mark.asyncio
    async def test_clear_expired_slots(self):
        """Should clear expired synchronization slots if no packets for 10 seconds."""
        with freeze_time(start_time) as frozen_datetime:
            await throttler.schedule_synchronize('accountId1', {'requestId': 'test1'})
            await throttler.schedule_synchronize('accountId2', {'requestId': 'test2'})
            asyncio.create_task(throttler.schedule_synchronize('accountId3', {'requestId': 'test3'}))
            await sleep(0.2)
            assert client._rpc_request.call_count == 2
            frozen_datetime.tick(20)
            await sleep(0.2)
            assert client._rpc_request.call_count == 3

    @pytest.mark.asyncio
    async def test_renew_sync(self):
        """Should renew sync on update."""
        with freeze_time(start_time) as frozen_datetime:
            await throttler.schedule_synchronize('accountId1', {'requestId': 'test1'})
            await throttler.schedule_synchronize('accountId2', {'requestId': 'test2'})
            asyncio.create_task(throttler.schedule_synchronize('accountId3', {'requestId': 'test3'}))
            await sleep(0.2)
            assert client._rpc_request.call_count == 2
            frozen_datetime.tick(11)
            await sleep(0.2)
            assert client._rpc_request.call_count == 3
            frozen_datetime.tick(11)
            throttler.update_synchronization_id('test1')
            asyncio.create_task(throttler.schedule_synchronize('accountId4', {'requestId': 'test4'}))
            asyncio.create_task(throttler.schedule_synchronize('accountId5', {'requestId': 'test5'}))
            await sleep(0.2)
            assert client._rpc_request.call_count == 4

    @pytest.mark.asyncio
    async def test_replace_previous_syncs(self):
        """Should replace previous syncs."""
        asyncio.create_task(throttler.schedule_synchronize('accountId1', {'requestId': 'test1'}))
        asyncio.create_task(throttler.schedule_synchronize('accountId1', {'requestId': 'test2'}))
        asyncio.create_task(throttler.schedule_synchronize('accountId1', {'requestId': 'test3'}))
        asyncio.create_task(throttler.schedule_synchronize('accountId2', {'requestId': 'test4'}))
        asyncio.create_task(throttler.schedule_synchronize('accountId3', {'requestId': 'test5'}))
        asyncio.create_task(throttler.schedule_synchronize('accountId1', {'requestId': 'test6', 'instanceIndex': 0}))
        await sleep(0.2)
        assert client._rpc_request.call_count == 4

    @pytest.mark.asyncio
    async def test_clear_on_disconnect(self):
        """Should clear existing sync ids on disconnect."""
        await throttler.schedule_synchronize('accountId1', {'requestId': 'test1'})
        await throttler.schedule_synchronize('accountId2', {'requestId': 'test2'})
        asyncio.create_task(throttler.schedule_synchronize('accountId3', {'requestId': 'test3'}))
        await sleep(0.2)
        assert client._rpc_request.call_count == 2
        throttler.on_disconnect()
        await sleep(0.2)
        assert client._rpc_request.call_count == 3

    @pytest.mark.asyncio
    async def test_remove_from_queue(self):
        """Should remove synchronizations from queue."""
        with freeze_time(start_time) as frozen_datetime:
            await throttler.schedule_synchronize('accountId1', {'requestId': 'test1'})
            await throttler.schedule_synchronize('accountId2', {'requestId': 'test2'})
            asyncio.create_task(throttler.schedule_synchronize('accountId3', {'requestId': 'test3'}))
            asyncio.create_task(throttler.schedule_synchronize('accountId3', {'requestId': 'test4',
                                                                              'instanceIndex': 0}))
            asyncio.create_task(throttler.schedule_synchronize('accountId4', {'requestId': 'test5'}))
            asyncio.create_task(throttler.schedule_synchronize('accountId3', {'requestId': 'test6'}))
            asyncio.create_task(throttler.schedule_synchronize('accountId4', {'requestId': 'test7'}))
            asyncio.create_task(throttler.schedule_synchronize('accountId3', {'requestId': 'test8'}))
            asyncio.create_task(throttler.schedule_synchronize('accountId5', {'requestId': 'test9'}))
            asyncio.create_task(throttler.schedule_synchronize('accountId3', {'requestId': 'test10',
                                                                              'instanceIndex': 0}))
            await sleep(0.2)
            frozen_datetime.tick(21)
            await sleep(0.2)
            frozen_datetime.tick(21)
            await sleep(0.2)
            frozen_datetime.tick(21)
            await sleep(0.2)
            frozen_datetime.tick(21)
            await sleep(0.2)
            assert client._rpc_request.call_count == 6
            client._rpc_request.assert_any_call('accountId1', {'requestId': 'test1'})
            client._rpc_request.assert_any_call('accountId2', {'requestId': 'test2'})
            client._rpc_request.assert_any_call('accountId3', {'requestId': 'test8'})
            client._rpc_request.assert_any_call('accountId3', {'requestId': 'test10', 'instanceIndex': 0})
            client._rpc_request.assert_any_call('accountId4', {'requestId': 'test7'})
            client._rpc_request.assert_any_call('accountId5', {'requestId': 'test9'})

    @pytest.mark.asyncio
    async def test_remove_expired_from_queue(self):
        """Should remove expired synchronizations from queue."""
        with freeze_time(start_time) as frozen_datetime:
            await throttler.schedule_synchronize('accountId1', {'requestId': 'test1'})
            await throttler.schedule_synchronize('accountId2', {'requestId': 'test2'})
            asyncio.create_task(throttler.schedule_synchronize('accountId3', {'requestId': 'test3'}))
            asyncio.create_task(throttler.schedule_synchronize('accountId4', {'requestId': 'test4'}))
            await sleep(0.1)
            frozen_datetime.tick(160)
            asyncio.create_task(throttler.schedule_synchronize('accountId5', {'requestId': 'test5'}))
            frozen_datetime.tick(160)
            throttler.update_synchronization_id('test1')
            throttler.update_synchronization_id('test2')
            await sleep(0.1)
            frozen_datetime.tick(21)
            await sleep(0.1)
            frozen_datetime.tick(21)
            await sleep(0.1)
            assert client._rpc_request.call_count == 3
            client._rpc_request.assert_any_call('accountId1', {'requestId': 'test1'})
            client._rpc_request.assert_any_call('accountId2', {'requestId': 'test2'})
            client._rpc_request.assert_any_call('accountId5', {'requestId': 'test5'})



from datetime import datetime
from typing_extensions import TypedDict
from typing import List, Optional
import iso8601
import random
import string
import pytz
import re


def date(date_time: str or float or int) -> datetime:
    """Parses a date string into a datetime object."""
    if isinstance(date_time, float) or isinstance(date_time, int):
        return datetime.fromtimestamp(max(date_time, 100000)).astimezone(pytz.utc)
    else:
        return iso8601.parse_date(date_time)


def format_date(date: datetime) -> str:
    """Converts date to format compatible with JS"""
    return date.astimezone(pytz.utc).isoformat(timespec='milliseconds').replace('+00:00', 'Z')


def random_id(length: int = 32) -> str:
    """Generates a random id of 32 symbols."""
    return ''.join(random.choice(string.ascii_lowercase) for i in range(length))


def convert_iso_time_to_date(packet):
    for field in packet:
        value = packet[field]
        if isinstance(value, str) and re.search('time|Time', field) and not \
                re.search('brokerTime|BrokerTime', field):
            packet[field] = date(value)
        if isinstance(value, list):
            for item in value:
                convert_iso_time_to_date(item)
        if isinstance(value, dict):
            convert_iso_time_to_date(value)
    if packet and 'timestamps' in packet:
        for field in packet['timestamps']:
            packet['timestamps'][field] = date(packet['timestamps'][field])
    if packet and 'type' in packet and packet['type'] == 'prices':
        if 'prices' in packet:
            for price in packet['prices']:
                if 'timestamps' in price:
                    for field in price['timestamps']:
                        if isinstance(price['timestamps'][field], str):
                            price['timestamps'][field] = date(price['timestamps'][field])


class MetatraderAccountInformation(TypedDict):
    """MetaTrader account information (see https://metaapi.cloud/docs/client/models/metatraderAccountInformation/)"""

    platform: str
    """Platform id (mt4 or mt5)"""
    broker: str
    """Broker name."""
    currency: str
    """Account base currency ISO code."""
    server: str
    """Broker server name."""
    balance: float
    """Account balance."""
    equity: float
    """Account liquidation value."""
    margin: float
    """Used margin."""
    freeMargin: float
    """Free margin."""
    leverage: float
    """Account leverage coefficient."""
    marginLevel: float
    """Margin level calculated as % of equity/margin."""
    tradeAllowed: bool
    """Flag indicating that trading is allowed."""
    investorMode: Optional[bool]
    """Flag indicating that investor password was used (supported for g2 only)."""
    marginMode: str
    """Margin calculation mode, one of ACCOUNT_MARGIN_MODE_EXCHANGE, ACCOUNT_MARGIN_MODE_RETAIL_NETTING,
    ACCOUNT_MARGIN_MODE_RETAIL_HEDGING."""
    name: str
    """Account owner name."""
    login: int
    """Account login."""
    credit: float
    """Account credit in the deposit currency."""


class MetatraderPosition(TypedDict):
    """MetaTrader position"""

    id: int
    """Position id (ticket number)."""
    type: str
    """Position type (one of POSITION_TYPE_BUY, POSITION_TYPE_SELL)."""
    symbol: str
    """Position symbol."""
    magic: int
    """Position magic number, identifies the EA which opened the position."""
    time: datetime
    """Time position was opened at."""
    brokerTime: str
    """Time position was opened at, in broker timezone, YYYY-MM-DD HH:mm:ss.SSS format."""
    updateTime: datetime
    """Last position modification time."""
    openPrice: float
    """Position open price."""
    currentPrice: float
    """Current price."""
    currentTickValue: float
    """Current tick value."""
    stopLoss: Optional[float]
    """Optional position stop loss price."""
    takeProfit: Optional[float]
    """Optional position take profit price."""
    volume: float
    """Position volume."""
    swap: float
    """Position cumulative swap."""
    profit: float
    """Position cumulative profit."""
    comment: Optional[str]
    """Optional position comment. The sum of the line lengths of the comment and the clientId
    must be less than or equal to 26. For more information see https://metaapi.cloud/docs/client/clientIdUsage/"""
    clientId: Optional[str]
    """Optional client-assigned id. The id value can be assigned when submitting a trade and
    will be present on position, history orders and history deals related to the trade. You can use this field to bind
    your trades to objects in your application and then track trade progress. The sum of the line lengths of the
    comment and the clientId must be less than or equal to 26. For more information see
    https://metaapi.cloud/docs/client/clientIdUsage/"""
    unrealizedProfit: float
    """Profit of the part of the position which is not yet closed, including swap."""
    realizedProfit: float
    """Profit of the already closed part, including commissions and swap."""
    commission: Optional[float]
    """Optional position commission."""
    reason: str
    """Position opening reason. One of POSITION_REASON_CLIENT, POSITION_REASON_EXPERT, POSITION_REASON_MOBILE,
    POSITION_REASON_WEB, POSITION_REASON_UNKNOWN. See
    https://www.mql5.com/en/docs/constants/tradingconstants/positionproperties#enum_position_reason"""
    accountCurrencyExchangeRate: Optional[float]
    """Current exchange rate of account currency into USD."""
    originalComment: Optional[str]
    """Position original comment (present if possible to restore from history)."""
    updatePending: Optional[bool]
    """Flag indicating that position original comment and clientId was not identified yet and will be
    updated in a future packet"""


class MetatraderOrder(TypedDict):
    """MetaTrader order"""

    id: int
    """Order id (ticket number)."""
    type: str
    """Order type (one of ORDER_TYPE_SELL, ORDER_TYPE_BUY, ORDER_TYPE_BUY_LIMIT,
    ORDER_TYPE_SELL_LIMIT, ORDER_TYPE_BUY_STOP, ORDER_TYPE_SELL_STOP). See
    https://www.mql5.com/en/docs/constants/tradingconstants/orderproperties#enum_order_type"""
    state: str
    """Order state one of (ORDER_STATE_STARTED, ORDER_STATE_PLACED, ORDER_STATE_CANCELED,
    ORDER_STATE_PARTIAL, ORDER_STATE_FILLED, ORDER_STATE_REJECTED, ORDER_STATE_EXPIRED, ORDER_STATE_REQUEST_ADD,
    ORDER_STATE_REQUEST_MODIFY, ORDER_STATE_REQUEST_CANCEL). See
    https://www.mql5.com/en/docs/constants/tradingconstants/orderproperties#enum_order_state"""
    magic: int
    """Order magic number, identifies the EA which created the order."""
    time: datetime
    """Time order was created at."""
    brokerTime: str
    """Time position was opened at, in broker timezone, YYYY-MM-DD HH:mm:ss.SSS format."""
    doneTime: Optional[datetime]
    """Optional time order was executed or canceled at. Will be specified for completed orders only."""
    doneBrokerTime: Optional[str]
    """Time order was executed or canceled at, in broker timezone, YYYY-MM-DD HH:mm:ss.SSS format. Will be specified
    for completed orders only"""
    symbol: str
    """Order symbol."""
    openPrice: float
    """Order open price (market price for market orders, limit price for limit orders or stop price for stop orders)."""
    currentPrice: float
    """Current price."""
    stopLoss: Optional[float]
    """Optional order stop loss price."""
    takeProfit: Optional[float]
    """Optional order take profit price."""
    volume: float
    """Order requested quantity."""
    currentVolume: float
    """Order remaining quantity, i.e. requested quantity - filled quantity."""
    positionId: str
    """Order position id. Present only if the order has a position attached to it."""
    comment: Optional[str]
    """Optional order comment. The sum of the line lengths of the comment and the clientId
    must be less than or equal to 26. For more information see https://metaapi.cloud/docs/client/clientIdUsage/"""
    originalComment: Optional[str]
    """Optional order original comment (present if possible to restore original comment from history)"""
    clientId: Optional[str]
    """Optional client-assigned id. The id value can be assigned when submitting a trade and
    will be present on position, history orders and history deals related to the trade. You can use this field to bind
    your trades to objects in your application and then track trade progress. The sum of the line lengths of the
    comment and the clientId must be less than or equal to 26. For more information see
    https://metaapi.cloud/docs/client/clientIdUsage/"""
    platform: str
    """Platform id (mt4 or mt5)."""
    updatePending: Optional[bool]
    """Optional flag indicating that order client id and original comment was not
    identified yet and will be updated in a future synchronization packet."""
    reason: str
    """Order opening reason. One of ORDER_REASON_CLIENT, ORDER_REASON_MOBILE, ORDER_REASON_WEB,
    ORDER_REASON_EXPERT, ORDER_REASON_SL, ORDER_REASON_TP, ORDER_REASON_SO, ORDER_REASON_UNKNOWN. See
    https://www.mql5.com/en/docs/constants/tradingconstants/orderproperties#enum_order_reason."""
    fillingMode: str
    """Order filling mode. One of ORDER_FILLING_FOK, ORDER_FILLING_IOC, ORDER_FILLING_RETURN. See
    https://www.mql5.com/en/docs/constants/tradingconstants/orderproperties#enum_order_type_filling."""
    expirationType: str
    """Order expiration type. One of ORDER_TIME_GTC, ORDER_TIME_DAY, ORDER_TIME_SPECIFIED, ORDER_TIME_SPECIFIED_DAY.
    See https://www.mql5.com/en/docs/constants/tradingconstants/orderproperties#enum_order_type_time"""
    expirationTime: datetime
    """Optional order expiration time."""
    accountCurrencyExchangeRate: Optional[float]
    """Current exchange rate of account currency into USD."""
    closeByPositionId: Optional[str]
    """Identifier of an opposite position used for closing by order ORDER_TYPE_CLOSE_BY"""
    stopLimitPrice: Optional[float]
    """The limit order price for the StopLimit order."""


class MetatraderHistoryOrders(TypedDict):
    """MetaTrader history orders search query response."""

    historyOrders: List[MetatraderOrder]
    """Array of history orders returned."""
    synchronizing: bool
    """Flag indicating that history order initial synchronization is still in progress
    and thus search results may be incomplete"""


class MetatraderDeal(TypedDict):
    """MetaTrader deal"""

    id: str
    """Deal id (ticket number)"""
    type: str
    """Deal type (one of DEAL_TYPE_BUY, DEAL_TYPE_SELL, DEAL_TYPE_BALANCE, DEAL_TYPE_CREDIT,
    DEAL_TYPE_CHARGE, DEAL_TYPE_CORRECTION, DEAL_TYPE_BONUS, DEAL_TYPE_COMMISSION, DEAL_TYPE_COMMISSION_DAILY,
    DEAL_TYPE_COMMISSION_MONTHLY, DEAL_TYPE_COMMISSION_AGENT_DAILY, DEAL_TYPE_COMMISSION_AGENT_MONTHLY,
    DEAL_TYPE_INTEREST, DEAL_TYPE_BUY_CANCELED, DEAL_TYPE_SELL_CANCELED, DEAL_DIVIDEND, DEAL_DIVIDEND_FRANKED,
    DEAL_TAX). See https://www.mql5.com/en/docs/constants/tradingconstants/dealproperties#enum_deal_type"""
    entryType: str
    """Deal entry type (one of DEAL_ENTRY_IN, DEAL_ENTRY_OUT, DEAL_ENTRY_INOUT,
    DEAL_ENTRY_OUT_BY). See https://www.mql5.com/en/docs/constants/tradingconstants/dealproperties#enum_deal_entry"""
    symbol: Optional[str]
    """Optional symbol deal relates to."""
    magic: Optional[int]
    """Optional deal magic number, identifies the EA which initiated the deal."""
    time: datetime
    """Time the deal was conducted at."""
    brokerTime: str
    """Time the deal was conducted at, in broker timezone, YYYY-MM-DD HH:mm:ss.SSS format."""
    volume: Optional[float]
    """Optional deal volume."""
    price: Optional[float]
    """Optional, the price the deal was conducted at."""
    commission: Optional[float]
    """Optional deal commission."""
    swap: Optional[float]
    """Optional deal swap."""
    profit: float
    """Deal profit."""
    positionId: Optional[str]
    """Optional id of position the deal relates to."""
    orderId: Optional[str]
    """Optional id of order the deal relates to."""
    comment: Optional[str]
    """Optional deal comment. The sum of the line lengths of the comment and the clientId
    must be less than or equal to 26. For more information see https://metaapi.cloud/docs/client/clientIdUsage/"""
    originalComment: Optional[str]
    """Optional deal original comment (present if possible to restore original comment from history)."""
    clientId: Optional[str]
    """Optional client-assigned id. The id value can be assigned when submitting a trade and will be present on
    position, history orders and history deals related to the trade. You can use this field to bind your trades
    to objects in your application and then track trade progress. The sum of the line lengths of the comment and
    the clientId must be less than or equal to 26. For more information see
    https://metaapi.cloud/docs/client/clientIdUsage/"""
    platform: str
    """Platform id (mt4 or mt5)."""
    updatePending: Optional[bool]
    """Optional flag indicating that deal client id and original comment was not
    identified yet and will be updated in a future synchronization packet"""
    reason: Optional[str]
    """Optional deal execution reason. One of DEAL_REASON_CLIENT, DEAL_REASON_MOBILE, DEAL_REASON_WEB,
    DEAL_REASON_EXPERT, DEAL_REASON_SL, DEAL_REASON_TP, DEAL_REASON_SO, DEAL_REASON_ROLLOVER, DEAL_REASON_VMARGIN,
    DEAL_REASON_SPLIT, DEAL_REASON_UNKNOWN. See
    https://www.mql5.com/en/docs/constants/tradingconstants/dealproperties#enum_deal_reason."""
    accountCurrencyExchangeRate: Optional[float]
    """Current exchange rate of account currency into USD."""


class MetatraderDeals(TypedDict):
    """MetaTrader history deals search query response"""

    deals: List[MetatraderDeal]
    """Array of history deals returned."""
    synchronizing: bool
    """Flag indicating that deal initial synchronization is still in progress
    and thus search results may be incomplete."""


MetatraderSession = TypedDict(
    "MetatraderSession",
    {
        "from": str,  # Session start time, in hh.mm.ss.SSS format.
        "to": str  # Session end time, in hh.mm.ss.SSS format.
    }
)
"""Metatrader trade or quote session"""


class MetatraderSessions(TypedDict):
    """Metatrader trade or quote session container, indexed by weekday."""
    SUNDAY: Optional[List[MetatraderSession]]
    """Array of sessions for SUNDAY."""
    MONDAY: Optional[List[MetatraderSession]]
    """Array of sessions for MONDAY."""
    TUESDAY: Optional[List[MetatraderSession]]
    """Array of sessions for TUESDAY."""
    WEDNESDAY: Optional[List[MetatraderSession]]
    """Array of sessions for WEDNESDAY."""
    THURSDAY: Optional[List[MetatraderSession]]
    """Array of sessions for THURSDAY."""
    FRIDAY: Optional[List[MetatraderSession]]
    """Array of sessions for FRIDAY."""
    SATURDAY: Optional[List[MetatraderSession]]
    """Array of sessions for SATURDAY."""


class MetatraderSymbolSpecification(TypedDict):
    """MetaTrader symbol specification. Contains symbol specification (see
    https://metaapi.cloud/docs/client/models/metatraderSymbolSpecification/)"""

    symbol: str
    """Symbol (e.g. a currency pair or an index)."""
    tickSize: float
    """Tick size"""
    minVolume: float
    """Minimum order volume for the symbol."""
    maxVolume: float
    """Maximum order volume for the symbol."""
    volumeStep: float
    """Order volume step for the symbol."""
    fillingModes: List[str]
    """List of allowed order filling modes. Can contain ORDER_FILLING_FOK, ORDER_FILLING_IOC or both.
    See https://www.mql5.com/en/docs/constants/environment_state/marketinfoconstants#symbol_filling_mode for more
    details."""
    executionMode: str
    """Deal execution mode. Possible values are SYMBOL_TRADE_EXECUTION_REQUEST, SYMBOL_TRADE_EXECUTION_INSTANT,
    SYMBOL_TRADE_EXECUTION_MARKET, SYMBOL_TRADE_EXECUTION_EXCHANGE. See
    https://www.mql5.com/en/docs/constants/environment_state/marketinfoconstants#enum_symbol_trade_execution for more
    details."""
    contractSize: float
    """Trade contract size."""
    quoteSessions: MetatraderSessions
    """Quote sessions, indexed by day of week."""
    tradeSessions: MetatraderSessions
    """Trade sessions, indexed by day of week."""
    tradeMode: Optional[str]
    """Order execution type. Possible values are SYMBOL_TRADE_MODE_DISABLED, SYMBOL_TRADE_MODE_LONGONLY,
    SYMBOL_TRADE_MODE_SHORTONLY, SYMBOL_TRADE_MODE_CLOSEONLY, SYMBOL_TRADE_MODE_FULL. See
    https://www.mql5.com/en/docs/constants/environment_state/marketinfoconstants#enum_symbol_trade_mode for more
    details."""
    bondAccruedInterest: Optional[float]
    """Accrued interest – accumulated coupon interest, i.e. part of the coupon interest calculated in proportion to
    the number of days since the coupon bond issuance or the last coupon interest payment."""
    bondFaceValue: Optional[float]
    """Face value – initial bond value set by the issuer."""
    optionStrike: Optional[float]
    """The strike price of an option. The price at which an option buyer can buy (in a Call option) or sell (in a
    Put option) the underlying asset, and the option seller is obliged to sell or buy the appropriate amount of the
    underlying asset."""
    optionPriceSensivity: Optional[float]
    """Option/warrant sensitivity shows by how many points the price of the option's underlying asset should change so
    that the price of the option changes by one point."""
    liquidityRate: Optional[float]
    """Liquidity Rate is the share of the asset that can be used for the margin."""
    initialMargin: float
    """Initial margin means the amount in the margin currency required for opening a position with the volume of one
    lot. It is used for checking a client's assets when he or she enters the market."""
    maintenanceMargin: float
    """The maintenance margin. If it is set, it sets the margin amount in the margin currency of the symbol, charged
    from one lot. It is used for checking a client's assets when his/her account state changes. If the maintenance
    margin is equal to 0, the initial margin is used."""
    hedgedMargin: float
    """Contract size or margin value per one lot of hedged positions (oppositely directed positions of one symbol).
    Two margin calculation methods are possible for hedged positions. The calculation method is defined by the broker"""
    hedgedMarginUsesLargerLeg: Optional[bool]
    """Calculating hedging margin using the larger leg (Buy or Sell)."""
    marginCurrency: str
    """Margin currency."""
    priceCalculationMode: str
    """Contract price calculation mode. One of SYMBOL_CALC_MODE_UNKNOWN, SYMBOL_CALC_MODE_FOREX,
    SYMBOL_CALC_MODE_FOREX_NO_LEVERAGE, SYMBOL_CALC_MODE_FUTURES, SYMBOL_CALC_MODE_CFD, SYMBOL_CALC_MODE_CFDINDEX,
    SYMBOL_CALC_MODE_CFDLEVERAGE, SYMBOL_CALC_MODE_EXCH_STOCKS, SYMBOL_CALC_MODE_EXCH_FUTURES,
    SYMBOL_CALC_MODE_EXCH_FUTURES_FORTS, SYMBOL_CALC_MODE_EXCH_BONDS, SYMBOL_CALC_MODE_EXCH_STOCKS_MOEX,
    SYMBOL_CALC_MODE_EXCH_BONDS_MOEX, SYMBOL_CALC_MODE_SERV_COLLATERAL. See
    https://www.mql5.com/en/docs/constants/environment_state/marketinfoconstants#enum_symbol_calc_mode for more
    details."""
    baseCurrency: str
    """Base currency."""
    profitCurrency: Optional[str]
    """Profit currency."""
    swapMode: str
    """Swap calculation model. Allowed values are SYMBOL_SWAP_MODE_DISABLED,
    SYMBOL_SWAP_MODE_POINTS, SYMBOL_SWAP_MODE_CURRENCY_SYMBOL, SYMBOL_SWAP_MODE_CURRENCY_MARGIN,
    SYMBOL_SWAP_MODE_CURRENCY_DEPOSIT, SYMBOL_SWAP_MODE_INTEREST_CURRENT, SYMBOL_SWAP_MODE_INTEREST_OPEN,
    SYMBOL_SWAP_MODE_REOPEN_CURRENT, SYMBOL_SWAP_MODE_REOPEN_BID. See
    https://www.mql5.com/en/docs/constants/environment_state/marketinfoconstants#enum_symbol_swap_mode for more
    details."""
    swapLong: Optional[float]
    """Long swap value."""
    swapShort: Optional[float]
    """Short swap value."""
    swapRollover3Days: str
    """Day of week to charge 3 days swap rollover. Allowed values are SUNDAY, MONDAY, TUESDAY, WEDNESDAY, THURDAY,
    FRIDAY, SATURDAY."""
    allowedExpirationModes: List[str]
    """Allowed order expiration modes. Allowed values are SYMBOL_EXPIRATION_GTC, SYMBOL_EXPIRATION_DAY,
    SYMBOL_EXPIRATION_SPECIFIED, SYMBOL_EXPIRATION_SPECIFIED_DAY.
    See https://www.mql5.com/en/docs/constants/environment_state/marketinfoconstants#symbol_expiration_mode for more
    details."""
    allowedOrderTypes: List[str]
    """Allowed order types. Allowed values are SYMBOL_ORDER_MARKET, SYMBOL_ORDER_LIMIT, SYMBOL_ORDER_STOP,
    SYMBOL_ORDER_STOP_LIMIT, SYMBOL_ORDER_SL, SYMBOL_ORDER_TP, SYMBOL_ORDER_CLOSEBY. See
    https://www.mql5.com/en/docs/constants/environment_state/marketinfoconstants#symbol_order_mode for more details."""
    orderGTCMode: str
    """If the expirationMode property is set to SYMBOL_EXPIRATION_GTC (good till canceled), the expiration of pending
    orders, as well as of Stop Loss/Take Profit orders should be additionally set using this enumeration. Allowed
    values are SYMBOL_ORDERS_GTC, SYMBOL_ORDERS_DAILY, SYMBOL_ORDERS_DAILY_EXCLUDING_STOPS. See
    https://www.mql5.com/en/docs/constants/environment_state/marketinfoconstants#enum_symbol_order_gtc_mode for more
    details."""
    digits: int
    """Digits after a decimal point."""
    path: Optional[str]
    """Path in the symbol tree."""
    description: str
    """Symbol description."""
    startTime: Optional[datetime]
    """Date of the symbol trade beginning (usually used for futures)."""
    expirationTime: Optional[datetime]
    """Date of the symbol trade end (usually used for futures)."""


class MetatraderSymbolPrice(TypedDict):
    """MetaTrader symbol price. Contains current price for a symbol (see
    https://metaapi.cloud/docs/client/models/metatraderSymbolPrice/)"""

    symbol: str
    """Symbol (e.g. a currency pair or an index)."""
    bid: float
    """Bid price."""
    ask: float
    """Ask price."""
    profitTickValue: float
    """Tick value for a profitable position."""
    lossTickValue: float
    """Tick value for a loosing position."""
    accountCurrencyExchangeRate: float
    """Current exchange rate of account currency into USD."""
    time: str
    """Quote time, in ISO format"""
    brokerTime: str
    """Quote time, in broker timezone, YYYY-MM-DD HH:mm:ss.SSS format."""


class MetatraderTradeResponse(TypedDict):
    """MetaTrader trade response."""

    numericCode: int
    """Numeric response code, see https://www.mql5.com/en/docs/constants/errorswarnings/enum_trade_return_codes and
    https://book.mql4.com/appendix/errors. Response codes which indicate success are 0, 10008-10010, 10025. The rest
    codes are errors."""
    stringCode: str
    """String response code, see https://www.mql5.com/en/docs/constants/errorswarnings/enum_trade_return_codes and
    https://book.mql4.com/appendix/errors. Response codes which indicate success are ERR_NO_ERROR,
    TRADE_RETCODE_PLACED, TRADE_RETCODE_DONE, TRADE_RETCODE_DONE_PARTIAL, TRADE_RETCODE_NO_CHANGES. The rest codes are
    errors."""
    message: str
    """Human-readable response message."""
    orderId: str
    """Order id which was created/modified during the trade."""
    positionId: str
    """Position id which was modified during the trade."""


class TradeOptions(TypedDict):
    """Common trade options."""

    comment: Optional[str]
    """Optional order comment. The sum of the line lengths of the comment and the clientId must be less than or equal
    to 26. For more information see https://metaapi.cloud/docs/client/clientIdUsage/"""
    clientId: Optional[str]
    """Optional client-assigned id. The id value can be assigned when submitting a trade and will be present on
    position, history orders and history deals related to the trade. You can use this field to bind your trades to
    objects in your application and then track trade progress. The sum of the line lengths of the comment and the
    clientId must be less than or equal to 26. For more information see
    https://metaapi.cloud/docs/client/clientIdUsage/"""
    magic: Optional[str]
    """Magic (expert id) number. If not set default value specified in account entity will be used."""
    slippage: Optional[int]
    """Optional slippage in points. Should be greater or equal to zero. In not set, default value specified in
    account entity will be used. Slippage is ignored if execution mode set to SYMBOL_TRADE_EXECUTION_MARKET in
    symbol specification. Not used for close by orders."""


class MarketTradeOptions(TradeOptions):
    """Market trade options."""

    fillingModes: Optional[List[str]]
    """Optional allowed filling modes in the order of priority. Default is to allow all filling modes and prefer
    ORDER_FILLING_FOK over ORDER_FILLING_IOC. See
    https://www.mql5.com/en/docs/constants/tradingconstants/orderproperties#enum_order_type_filling for extra
    explanation."""


class ExpirationOptions(TypedDict):
    """Pending order expiration settings."""

    type: str
    """Pending order expiration type. See
    https://www.mql5.com/en/docs/constants/tradingconstants/orderproperties#enum_order_type_time for the list of
    possible options. MetaTrader4 platform supports only ORDER_TIME_SPECIFIED expiration type. One of ORDER_TIME_GTC,
    ORDER_TIME_DAY, ORDER_TIME_SPECIFIED, ORDER_TIME_SPECIFIED_DAY."""
    time: Optional[datetime]
    """Optional pending order expiration time. Ignored if expiration type is not one of ORDER_TIME_DAY or
    ORDER_TIME_SPECIFIED."""


class PendingTradeOptions(TradeOptions):
    """Pending order trade options."""

    expiration: Optional[ExpirationOptions]
    """Optional pending order expiration settings. See Pending order expiration settings section."""


class ValidationDetails(TypedDict):
    """Object to supply additional information for validation exceptions."""
    parameter: str
    """Name of invalid parameter."""
    value: Optional[str]
    """Entered invalid value."""
    message: str
    """Error message."""


class ExceptionMessage(TypedDict):
    """A REST API response that contains an exception message"""
    id: int
    """Error id"""
    error: str
    """Error name"""
    numericCode: Optional[int]
    """Numeric error code"""
    stringCode: Optional[str]
    """String error code"""
    message: str
    """Human-readable error message"""
    details: Optional[List[ValidationDetails]]
    """Additional information about error. Used to supply validation error details."""



import pytest
import respx
from httpx import Response
from ..httpClient import HttpClient
from .metatraderDemoAccount_client import MetatraderDemoAccountClient
PROVISIONING_API_URL = 'https://mt-provisioning-api-v1.agiliumtrade.agiliumtrade.ai'
http_client = HttpClient()
demo_account_client = MetatraderDemoAccountClient(http_client, 'header.payload.sign')


class TestMetatraderDemoAccountClient:
    @respx.mock
    @pytest.mark.asyncio
    async def test_create_mt4(self):
        """Should create new MetaTrader 4 demo from API."""
        expected = {
            'login': '12345',
            'password': 'qwerty',
            'serverName': 'HugosWay-Demo3',
            'investorPassword': 'qwerty'
        }
        account = {
            'balance': 10,
            'email': 'test@test.com',
            'leverage': 15,
            'serverName': 'server'
        }
        rsps = respx.post(f'{PROVISIONING_API_URL}/users/current/provisioning-profiles/profileId1/mt4-demo-accounts') \
            .mock(return_value=Response(200, json=expected))
        accounts = await demo_account_client.create_mt4_demo_account('profileId1', account)
        assert rsps.calls[0].request.url == f'{PROVISIONING_API_URL}/users/current/provisioning-profiles/' + \
            'profileId1/mt4-demo-accounts'
        assert rsps.calls[0].request.method == 'POST'
        assert rsps.calls[0].request.headers['auth-token'] == 'header.payload.sign'
        assert accounts == expected

    @pytest.mark.asyncio
    async def test_not_create_mt4_demo_with_account_token(self):
        """Should not create MetaTrader 4 demo account via API with account token'."""
        account_client = MetatraderDemoAccountClient(http_client, 'token')
        try:
            await account_client.create_mt4_demo_account('', {})
        except Exception as err:
            assert err.__str__() == 'You can not invoke create_mt4_demo_account method, because you have ' + \
                                    'connected with account access token. Please use API access token from ' + \
                                    'https://app.metaapi.cloud/token page to invoke this method.'

    @respx.mock
    @pytest.mark.asyncio
    async def test_create_mt5(self):
        """Should create new MetaTrader 4 demo from API."""
        expected = {
            'login': '12345',
            'password': 'qwerty',
            'serverName': 'HugosWay-Demo3',
            'investorPassword': 'qwerty'
        }
        account = {
            'balance': 10,
            'email': 'test@test.com',
            'leverage': 15,
            'serverName': 'server'
        }
        rsps = respx.post(f'{PROVISIONING_API_URL}/users/current/provisioning-profiles/profileId2/mt5-demo-accounts') \
            .mock(return_value=Response(200, json=expected))
        accounts = await demo_account_client.create_mt5_demo_account('profileId2', account)
        assert rsps.calls[0].request.url == f'{PROVISIONING_API_URL}/users/current/provisioning-profiles/' + \
               'profileId2/mt5-demo-accounts'
        assert rsps.calls[0].request.method == 'POST'
        assert rsps.calls[0].request.headers['auth-token'] == 'header.payload.sign'
        assert accounts == expected

    @pytest.mark.asyncio
    async def test_not_create_mt5_demo_with_account_token(self):
        """Should not create MetaTrader 5 demo account via API with account token'."""
        account_client = MetatraderDemoAccountClient(http_client, 'token')
        try:
            await account_client.create_mt5_demo_account('', {})
        except Exception as err:
            assert err.__str__() == 'You can not invoke create_mt5_demo_account method, because you have ' + \
                                    'connected with account access token. Please use API access token from ' + \
                                    'https://app.metaapi.cloud/token page to invoke this method.'



from .historyStorage import HistoryStorage
from ..metaApi.metatraderAccountModel import MetatraderAccountModel
from abc import ABC, abstractmethod
from datetime import datetime


class ConnectionRegistryModel(ABC):
    """Defines interface for a connection registry class."""

    @abstractmethod
    async def connect(self, account: MetatraderAccountModel, history_storage: HistoryStorage,
                      history_start_time: datetime = None):
        """Creates and returns a new account connection if doesnt exist, otherwise returns old.

        Args:
            account: MetaTrader account to connect to.
            history_storage: Terminal history storage.
            history_start_time: History start time.

        Returns:
            A coroutine resolving with account connection.
        """

    @abstractmethod
    def remove(self, account_id: str):
        """Removes an account from registry.

        Args:
            account_id: MetaTrader account id to remove.
        """



class MethodAccessException(Exception):
    """Exception which indicates that user doesn't have access to a method.
    """

    def __init__(self, method_name: str, access_type: str = 'api'):
        """Inits the method access exception.

        Args:
            method_name: Name of method.
            access_type: Type of method access.
        """
        if access_type == 'api':
            error_message = f'You can not invoke {method_name} method, because you have connected with API ' + \
                            'access token. Please use account access token to invoke this method.'
        elif access_type == 'account':
            error_message = f'You can not invoke {method_name} method, because you have connected with account ' + \
                            'access token. Please use API access token from https://app.metaapi.cloud/token page ' + \
                            'to invoke this method.'
        else:
            error_message = ''
        super().__init__(error_message)



from ..clients.metaApi.synchronizationListener import SynchronizationListener
from ..clients.metaApi.reconnectListener import ReconnectListener
from ..clients.metaApi.metaApiWebsocket_client import MetaApiWebsocketClient
from .terminalState import TerminalState
from .connectionHealthMonitor import ConnectionHealthMonitor
from .memoryHistoryStorage import MemoryHistoryStorage
from .metatraderAccountModel import MetatraderAccountModel
from .connectionRegistryModel import ConnectionRegistryModel
from .historyStorage import HistoryStorage
from ..clients.timeoutException import TimeoutException
from .models import random_id, MetatraderSymbolSpecification, MetatraderAccountInformation, \
    MetatraderPosition, MetatraderOrder, MetatraderHistoryOrders, MetatraderDeals, MetatraderTradeResponse, \
    MetatraderSymbolPrice, MarketTradeOptions, PendingTradeOptions
from datetime import datetime, timedelta
from typing import Coroutine, List, Optional, Dict
from typing_extensions import TypedDict
from functools import reduce
import pytz
import asyncio


class MetaApiConnectionDict(TypedDict):
    instanceIndex: int
    ordersSynchronized: dict
    dealsSynchronized: dict
    shouldSynchronize: Optional[str]
    synchronizationRetryIntervalInSeconds: float
    synchronized: bool
    lastDisconnectedSynchronizationId: Optional[str]
    lastSynchronizationId: Optional[str]
    disconnected: bool


class SynchronizationOptions(TypedDict):
    instanceIndex: Optional[int]
    """Index of an account instance to ensure synchronization on, default is to wait for the first instance to
    synchronize."""
    applicationPattern: Optional[str]
    """Application regular expression pattern, default is .*"""
    synchronizationId: Optional[str]
    """synchronization id, last synchronization request id will be used by default"""
    timeoutInSeconds: Optional[float]
    """Wait timeout in seconds, default is 5m."""
    intervalInMilliseconds: Optional[float]
    """Interval between account reloads while waiting for a change, default is 1s."""


class MetaApiConnection(SynchronizationListener, ReconnectListener):
    """Exposes MetaApi MetaTrader API connection to consumers."""

    def __init__(self, websocket_client: MetaApiWebsocketClient, account: MetatraderAccountModel,
                 history_storage: HistoryStorage or None, connection_registry: ConnectionRegistryModel,
                 history_start_time: datetime = None):
        """Inits MetaApi MetaTrader Api connection.

        Args:
            websocket_client: MetaApi websocket client.
            account: MetaTrader account id to connect to.
            history_storage: Local terminal history storage. By default an instance of MemoryHistoryStorage
            will be used.
            history_start_time: History start sync time.
        """
        super().__init__()
        self._websocketClient = websocket_client
        self._account = account
        self._closed = False
        self._connection_registry = connection_registry
        self._history_start_time = history_start_time
        self._terminalState = TerminalState()
        self._historyStorage = history_storage or MemoryHistoryStorage(account.id)
        self._healthMonitor = ConnectionHealthMonitor(self)
        self._websocketClient.add_synchronization_listener(account.id, self)
        self._websocketClient.add_synchronization_listener(account.id, self._terminalState)
        self._websocketClient.add_synchronization_listener(account.id, self._historyStorage)
        self._websocketClient.add_synchronization_listener(account.id, self._healthMonitor)
        self._websocketClient.add_reconnect_listener(self)
        self._subscriptions = {}
        self._stateByInstanceIndex = {}
        self._synchronized = False
        self._shouldRetrySubscribe = False
        self._isSubscribing = False
        self._subscribeTask = None
        self._subscribeFuture = None

    def get_account_information(self) -> 'Coroutine[asyncio.Future[MetatraderAccountInformation]]':
        """Returns account information (see
        https://metaapi.cloud/docs/client/websocket/api/readTradingTerminalState/readAccountInformation/).

        Returns:
            A coroutine resolving with account information.
        """
        return self._websocketClient.get_account_information(self._account.id)

    def get_positions(self) -> 'Coroutine[asyncio.Future[List[MetatraderPosition]]]':
        """Returns positions (see
        https://metaapi.cloud/docs/client/websocket/api/readTradingTerminalState/readPositions/).

        Returns:
            A coroutine resolving with array of open positions.
        """
        return self._websocketClient.get_positions(self._account.id)

    def get_position(self, position_id: str) -> 'Coroutine[asyncio.Future[MetatraderPosition]]':
        """Returns specific position (see
        https://metaapi.cloud/docs/client/websocket/api/readTradingTerminalState/readPosition/).

        Args:
            position_id: Position id.

        Returns:
            A coroutine resolving with MetaTrader position found.
        """
        return self._websocketClient.get_position(self._account.id, position_id)

    def get_orders(self) -> 'Coroutine[asyncio.Future[List[MetatraderOrder]]]':
        """Returns open orders (see
        https://metaapi.cloud/docs/client/websocket/api/readTradingTerminalState/readOrders/).

        Returns:
            A coroutine resolving with open MetaTrader orders.
        """
        return self._websocketClient.get_orders(self._account.id)

    def get_order(self, order_id: str) -> 'Coroutine[asyncio.Future[MetatraderOrder]]':
        """Returns specific open order (see
        https://metaapi.cloud/docs/client/websocket/api/readTradingTerminalState/readOrder/).

        Args:
            order_id: Order id (ticket number).

        Returns:
            A coroutine resolving with metatrader order found.
        """
        return self._websocketClient.get_order(self._account.id, order_id)

    def get_history_orders_by_ticket(self, ticket: str) -> 'Coroutine[MetatraderHistoryOrders]':
        """Returns the history of completed orders for a specific ticket number (see
        https://metaapi.cloud/docs/client/websocket/api/retrieveHistoricalData/readHistoryOrdersByTicket/).

        Args:
            ticket: Ticket number (order id).

        Returns:
            A coroutine resolving with request results containing history orders found.
        """
        return self._websocketClient.get_history_orders_by_ticket(self._account.id, ticket)

    def get_history_orders_by_position(self, position_id: str) -> 'Coroutine[MetatraderHistoryOrders]':
        """Returns the history of completed orders for a specific position id (see
        https://metaapi.cloud/docs/client/websocket/api/retrieveHistoricalData/readHistoryOrdersByPosition/)

        Args:
            position_id: Position id.

        Returns:
            A coroutine resolving with request results containing history orders found.
        """
        return self._websocketClient.get_history_orders_by_position(self._account.id, position_id)

    def get_history_orders_by_time_range(self, start_time: datetime, end_time: datetime, offset: int = 0,
                                         limit: int = 1000) -> 'Coroutine[MetatraderHistoryOrders]':
        """Returns the history of completed orders for a specific time range (see
        https://metaapi.cloud/docs/client/websocket/api/retrieveHistoricalData/readHistoryOrdersByTimeRange/)

        Args:
            start_time: Start of time range, inclusive.
            end_time: End of time range, exclusive.
            offset: Pagination offset, default is 0.
            limit: Pagination limit, default is 1000.

        Returns:
            A coroutine resolving with request results containing history orders found.
        """
        return self._websocketClient.get_history_orders_by_time_range(self._account.id, start_time, end_time,
                                                                      offset, limit)

    def get_deals_by_ticket(self, ticket: str) -> 'Coroutine[MetatraderDeals]':
        """Returns history deals with a specific ticket number (see
        https://metaapi.cloud/docs/client/websocket/api/retrieveHistoricalData/readDealsByTicket/).

        Args:
            ticket: Ticket number (deal id for MT5 or order id for MT4).

        Returns:
            A coroutine resolving with request results containing deals found.
        """
        return self._websocketClient.get_deals_by_ticket(self._account.id, ticket)

    def get_deals_by_position(self, position_id) -> 'Coroutine[MetatraderDeals]':
        """Returns history deals for a specific position id (see
        https://metaapi.cloud/docs/client/websocket/api/retrieveHistoricalData/readDealsByPosition/).

        Args:
            position_id: Position id.

        Returns:
            A coroutine resolving with request results containing deals found.
        """
        return self._websocketClient.get_deals_by_position(self._account.id, position_id)

    def get_deals_by_time_range(self, start_time: datetime, end_time: datetime, offset: int = 0,
                                limit: int = 1000) -> 'Coroutine[MetatraderDeals]':
        """Returns history deals with for a specific time range (see
        https://metaapi.cloud/docs/client/websocket/api/retrieveHistoricalData/readDealsByTimeRange/).

        Args:
            start_time: Start of time range, inclusive.
            end_time: End of time range, exclusive.
            offset: Pagination offset, default is 0.
            limit: Pagination limit, default is 1000.

        Returns:
            A coroutine resolving with request results containing deals found.
        """
        return self._websocketClient.get_deals_by_time_range(self._account.id, start_time, end_time, offset, limit)

    def remove_history(self, application: str = None) -> Coroutine:
        """Clears the order and transaction history of a specified account so that it can be synchronized from scratch
        (see https://metaapi.cloud/docs/client/websocket/api/removeHistory/).

        Args:
            application: Application to remove history for.

        Returns:
            A coroutine resolving when the history is cleared.
        """
        asyncio.create_task(self._historyStorage.clear())
        return self._websocketClient.remove_history(self._account.id, application)

    def remove_application(self):
        """Clears the order and transaction history of a specified application and removes application (see
        https://metaapi.cloud/docs/client/websocket/api/removeApplication/).

        Returns:
            A coroutine resolving when the history is cleared and application is removed.
        """
        asyncio.create_task(self._historyStorage.clear())
        return self._websocketClient.remove_application(self._account.id)

    def create_market_buy_order(self, symbol: str, volume: float, stop_loss: float = None, take_profit: float = None,
                                options: MarketTradeOptions = None) -> \
            'Coroutine[asyncio.Future[MetatraderTradeResponse]]':
        """Creates a market buy order (see https://metaapi.cloud/docs/client/websocket/api/trade/).

        Args:
            symbol: Symbol to trade.
            volume: Order volume.
            stop_loss: Optional stop loss price.
            take_profit: Optional take profit price.
            options: Optional trade options.

        Returns:
            A coroutine resolving with trade result.

        Raises:
            TradeException: On trade error, check error properties for error code details.
        """
        trade_params = {'actionType': 'ORDER_TYPE_BUY', 'symbol': symbol, 'volume': volume}
        if stop_loss:
            trade_params['stopLoss'] = stop_loss
        if take_profit:
            trade_params['takeProfit'] = take_profit
        trade_params.update(options or {})
        return self._websocketClient.trade(self._account.id, trade_params)

    def create_market_sell_order(self, symbol: str, volume: float, stop_loss: float = None, take_profit: float = None,
                                 options: MarketTradeOptions = None) -> \
            'Coroutine[asyncio.Future[MetatraderTradeResponse]]':
        """Creates a market sell order (see https://metaapi.cloud/docs/client/websocket/api/trade/).

        Args:
            symbol: Symbol to trade.
            volume: Order volume.
            stop_loss: Optional stop loss price.
            take_profit: Optional take profit price.
            options: Optional trade options.

        Returns:
            A coroutine resolving with trade result.

        Raises:
            TradeException: On trade error, check error properties for error code details.
        """
        trade_params = {'actionType': 'ORDER_TYPE_SELL', 'symbol': symbol, 'volume': volume}
        if stop_loss:
            trade_params['stopLoss'] = stop_loss
        if take_profit:
            trade_params['takeProfit'] = take_profit
        trade_params.update(options or {})
        return self._websocketClient.trade(self._account.id, trade_params)

    def create_limit_buy_order(self, symbol: str, volume: float, open_price: float, stop_loss: float = None,
                               take_profit: float = None, options: PendingTradeOptions = None) -> \
            'Coroutine[asyncio.Future[MetatraderTradeResponse]]':
        """Creates a limit buy order (see https://metaapi.cloud/docs/client/websocket/api/trade/).

        Args:
            symbol: Symbol to trade.
            volume: Order volume.
            open_price: Order limit price.
            stop_loss: Optional stop loss price.
            take_profit: Optional take profit price.
            options: Optional trade options.

        Returns:
            A coroutine resolving with trade result.

        Raises:
            TradeException: On trade error, check error properties for error code details.
        """
        trade_params = {'actionType': 'ORDER_TYPE_BUY_LIMIT', 'symbol': symbol, 'volume': volume,
                        'openPrice': open_price}
        if stop_loss:
            trade_params['stopLoss'] = stop_loss
        if take_profit:
            trade_params['takeProfit'] = take_profit
        trade_params.update(options or {})
        return self._websocketClient.trade(self._account.id, trade_params)

    def create_limit_sell_order(self, symbol: str, volume: float, open_price: float, stop_loss: float = None,
                                take_profit: float = None, options: PendingTradeOptions = None) -> \
            'Coroutine[asyncio.Future[MetatraderTradeResponse]]':
        """Creates a limit sell order (see https://metaapi.cloud/docs/client/websocket/api/trade/).

        Args:
            symbol: Symbol to trade.
            volume: Order volume.
            open_price: Order limit price.
            stop_loss: Optional stop loss price.
            take_profit: Optional take profit price.
            options: Optional trade options.

        Returns:
            A coroutine resolving with trade result.

        Raises:
            TradeException: On trade error, check error properties for error code details.
        """
        trade_params = {'actionType': 'ORDER_TYPE_SELL_LIMIT', 'symbol': symbol, 'volume': volume,
                        'openPrice': open_price}
        if stop_loss:
            trade_params['stopLoss'] = stop_loss
        if take_profit:
            trade_params['takeProfit'] = take_profit
        trade_params.update(options or {})
        return self._websocketClient.trade(self._account.id, trade_params)

    def create_stop_buy_order(self, symbol: str, volume: float, open_price: float, stop_loss: float = None,
                              take_profit: float = None, options: PendingTradeOptions = None) -> \
            'Coroutine[asyncio.Future[MetatraderTradeResponse]]':
        """Creates a stop buy order (see https://metaapi.cloud/docs/client/websocket/api/trade/).

        Args:
            symbol: Symbol to trade.
            volume: Order volume.
            open_price: Order limit price.
            stop_loss: Optional stop loss price.
            take_profit: Optional take profit price.
            options: Optional trade options.

        Returns:
            A coroutine resolving with trade result.

        Raises:
            TradeException: On trade error, check error properties for error code details.
        """
        trade_params = {'actionType': 'ORDER_TYPE_BUY_STOP', 'symbol': symbol, 'volume': volume,
                        'openPrice': open_price}
        if stop_loss:
            trade_params['stopLoss'] = stop_loss
        if take_profit:
            trade_params['takeProfit'] = take_profit
        trade_params.update(options or {})
        return self._websocketClient.trade(self._account.id, trade_params)

    def create_stop_sell_order(self, symbol: str, volume: float, open_price: float, stop_loss: float = None,
                               take_profit: float = None, options: PendingTradeOptions = None) -> \
            'Coroutine[asyncio.Future[MetatraderTradeResponse]]':
        """Creates a stop sell order (see https://metaapi.cloud/docs/client/websocket/api/trade/).

        Args:
            symbol: Symbol to trade.
            volume: Order volume.
            open_price: Order limit price.
            stop_loss: Optional stop loss price.
            take_profit: Optional take profit price.
            options: Optional trade options.

        Returns:
            A coroutine resolving with trade result.

        Raises:
            TradeException: On trade error, check error properties for error code details.
        """
        trade_params = {'actionType': 'ORDER_TYPE_SELL_STOP', 'symbol': symbol, 'volume': volume,
                        'openPrice': open_price}
        if stop_loss:
            trade_params['stopLoss'] = stop_loss
        if take_profit:
            trade_params['takeProfit'] = take_profit
        trade_params.update(options or {})
        return self._websocketClient.trade(self._account.id, trade_params)

    def create_stop_limit_buy_order(self, symbol: str, volume: float, open_price: float, stop_limit_price: float,
                                    stop_loss: float = None, take_profit: float = None,
                                    options: PendingTradeOptions = None):
        """Creates a stop limit buy order (see https://metaapi.cloud/docs/client/websocket/api/trade/).

        Args:
            symbol: Symbol to trade.
            volume: Order volume.
            open_price: Order limit price.
            stop_limit_price: The limit order price for the stop limit order.
            stop_loss: Optional stop loss price.
            take_profit: Optional take profit price.
            options: Optional trade options.

        Returns:
            A coroutine resolving with trade result.

        Raises:
            TradeException: On trade error, check error properties for error code details.
        """
        trade_params = {'actionType': 'ORDER_TYPE_BUY_STOP_LIMIT', 'symbol': symbol, 'volume': volume,
                        'openPrice': open_price, 'stopLimitPrice': stop_limit_price}
        if stop_loss:
            trade_params['stopLoss'] = stop_loss
        if take_profit:
            trade_params['takeProfit'] = take_profit
        trade_params.update(options or {})
        return self._websocketClient.trade(self._account.id, trade_params)

    def create_stop_limit_sell_order(self, symbol: str, volume: float, open_price: float, stop_limit_price: float,
                                     stop_loss: float = None, take_profit: float = None,
                                     options: PendingTradeOptions = None):
        """Creates a stop limit sell order (see https://metaapi.cloud/docs/client/websocket/api/trade/).

        Args:
            symbol: Symbol to trade.
            volume: Order volume.
            open_price: Order limit price.
            stop_limit_price: The limit order price for the stop limit order.
            stop_loss: Optional stop loss price.
            take_profit: Optional take profit price.
            options: Optional trade options.

        Returns:
            A coroutine resolving with trade result.

        Raises:
            TradeException: On trade error, check error properties for error code details.
        """
        trade_params = {'actionType': 'ORDER_TYPE_SELL_STOP_LIMIT', 'symbol': symbol, 'volume': volume,
                        'openPrice': open_price, 'stopLimitPrice': stop_limit_price}
        if stop_loss:
            trade_params['stopLoss'] = stop_loss
        if take_profit:
            trade_params['takeProfit'] = take_profit
        trade_params.update(options or {})
        return self._websocketClient.trade(self._account.id, trade_params)

    def modify_position(self, position_id: str, stop_loss: float = None, take_profit: float = None) -> \
            'Coroutine[asyncio.Future[MetatraderTradeResponse]]':
        """Modifies a position (see https://metaapi.cloud/docs/client/websocket/api/trade/).

        Args:
            position_id: Position id to modify.
            stop_loss: Optional stop loss price.
            take_profit: Optional take profit price.

        Returns:
            A coroutine resolving with trade result.

        Raises:
            TradeException: On trade error.
        """
        trade_params = {'actionType': 'POSITION_MODIFY', 'positionId': position_id}
        if stop_loss:
            trade_params['stopLoss'] = stop_loss
        if take_profit:
            trade_params['takeProfit'] = take_profit
        return self._websocketClient.trade(self._account.id, trade_params)

    def close_position_partially(self, position_id: str, volume: float, options: MarketTradeOptions = None) -> \
            'Coroutine[asyncio.Future[MetatraderTradeResponse]]':
        """Partially closes a position (see https://metaapi.cloud/docs/client/websocket/api/trade/).

        Args:
            position_id: Position id to modify.
            volume: Volume to close.
            options: Optional trade options.

        Returns:
            A coroutine resolving with trade result.

        Raises:
            TradeException: On trade error, check error properties for error code details.
        """
        trade_params = {'actionType': 'POSITION_PARTIAL', 'positionId': position_id, 'volume': volume}
        trade_params.update(options or {})
        return self._websocketClient.trade(self._account.id, trade_params)

    def close_position(self, position_id: str, options: MarketTradeOptions = None) -> \
            'Coroutine[asyncio.Future[MetatraderTradeResponse]]':
        """Fully closes a position (see https://metaapi.cloud/docs/client/websocket/api/trade/).

        Args:
            position_id: Position id to modify.
            options: Optional trade options.

        Returns:
            A coroutine resolving with trade result.

        Raises:
            TradeException: On trade error, check error properties for error code details.
        """
        trade_params = {'actionType': 'POSITION_CLOSE_ID', 'positionId': position_id}
        trade_params.update(options or {})
        return self._websocketClient.trade(self._account.id, trade_params)

    def close_by(self, position_id: str, opposite_position_id: str, options: MarketTradeOptions = None) -> \
            'Coroutine[asyncio.Future[MetatraderTradeResponse]]':
        """Fully closes a position (see https://metaapi.cloud/docs/client/websocket/api/trade/).

        Args:
            position_id: Position id to close by opposite position.
            opposite_position_id: Opposite position id to close.
            options: Optional trade options.

        Returns:
            A coroutine resolving with trade result.

        Raises:
            TradeException: On trade error, check error properties for error code details.
        """
        trade_params = {'actionType': 'POSITION_CLOSE_BY', 'positionId': position_id,
                        'closeByPositionId': opposite_position_id}
        trade_params.update(options or {})
        return self._websocketClient.trade(self._account.id, trade_params)

    def close_positions_by_symbol(self, symbol: str, options: MarketTradeOptions = None) -> \
            'Coroutine[asyncio.Future[MetatraderTradeResponse]]':
        """Closes positions by a symbol (see https://metaapi.cloud/docs/client/websocket/api/trade/).

        Args:
            symbol: Symbol to trade.
            options: Optional trade options.

        Returns:
            A coroutine resolving with trade result.

        Raises:
            TradeException: On trade error, check error properties for error code details.
        """
        trade_params = {'actionType': 'POSITIONS_CLOSE_SYMBOL', 'symbol': symbol}
        trade_params.update(options or {})
        return self._websocketClient.trade(self._account.id, trade_params)

    def modify_order(self, order_id: str, open_price: float, stop_loss: float = None,
                     take_profit: float = None) -> \
            'Coroutine[asyncio.Future[MetatraderTradeResponse]]':
        """Modifies a pending order (see https://metaapi.cloud/docs/client/websocket/api/trade/).

        Args:
            order_id: Order id (ticket number).
            open_price: Order stop price.
            stop_loss: Optional stop loss price.
            take_profit: Optional take profit price.

        Returns:
            A coroutine resolving with trade result.

        Raises:
            TradeException: On trade error, check error properties for error code details.
        """
        trade_params = {'actionType': 'ORDER_MODIFY', 'orderId': order_id, 'openPrice': open_price}
        if stop_loss:
            trade_params['stopLoss'] = stop_loss
        if take_profit:
            trade_params['takeProfit'] = take_profit
        return self._websocketClient.trade(self._account.id, trade_params)

    def cancel_order(self, order_id: str) -> \
            'Coroutine[asyncio.Future[MetatraderTradeResponse]]':
        """Cancels order (see https://metaapi.cloud/docs/client/websocket/api/trade/).

        Args:
            order_id: Order id (ticket number).

        Returns:
            A coroutine resolving with trade result.

        Raises:
            TradeException: On trade error, check error properties for error code details.
        """
        return self._websocketClient.trade(self._account.id, {'actionType': 'ORDER_CANCEL', 'orderId': order_id})

    def reconnect(self) -> Coroutine:
        """Reconnects to the Metatrader terminal (see https://metaapi.cloud/docs/client/websocket/api/reconnect/).

        Returns:
            A coroutine which resolves when reconnection started.
        """
        return self._websocketClient.reconnect(self._account.id)

    async def synchronize(self, instance_index: int) -> Coroutine:
        """Requests the terminal to start synchronization process.
        (see https://metaapi.cloud/docs/client/websocket/synchronizing/synchronize/).

        Args:
            instance_index: Instance index.

        Returns:
            A coroutine which resolves when synchronization started.
        """
        starting_history_order_time = \
            datetime.utcfromtimestamp(max(((self._history_start_time and self._history_start_time.timestamp()) or 0),
                                      (await self._historyStorage.last_history_order_time(instance_index))
                                          .timestamp())).replace(tzinfo=pytz.UTC)
        starting_deal_time = \
            datetime.utcfromtimestamp(max(((self._history_start_time and self._history_start_time.timestamp()) or 0),
                                      (await self._historyStorage.last_deal_time(instance_index)).timestamp()))\
            .replace(tzinfo=pytz.UTC)
        synchronization_id = random_id()
        self._get_state(instance_index)['lastSynchronizationId'] = synchronization_id
        return await self._websocketClient.synchronize(self._account.id, instance_index, synchronization_id,
                                                       starting_history_order_time, starting_deal_time)

    async def initialize(self):
        """Initializes meta api connection"""
        await self._historyStorage.initialize()

    async def subscribe(self):
        """Initiates subscription to MetaTrader terminal.

        Returns:
            A coroutine which resolves when subscription is initiated.
        """
        if not self._isSubscribing:
            self._isSubscribing = True
            self._shouldRetrySubscribe = True
            subscribe_retry_interval_in_seconds = 3
            while self._shouldRetrySubscribe and not self._closed:
                try:
                    await self._websocketClient.subscribe(self._account.id)
                except Exception:
                    pass
                if not self._shouldRetrySubscribe:
                    break
                retry_interval = subscribe_retry_interval_in_seconds
                subscribe_retry_interval_in_seconds = min(subscribe_retry_interval_in_seconds * 2, 300)
                subscribe_future = asyncio.Future()

                async def subscribe_task():
                    await asyncio.sleep(retry_interval)
                    subscribe_future.set_result(True)
                self._subscribeTask = asyncio.create_task(subscribe_task())
                self._subscribeFuture = subscribe_future
                result = await self._subscribeFuture
                self._subscribeFuture = None
                if not result:
                    break
            self._isSubscribing = False

    def subscribe_to_market_data(self, symbol: str, instance_index: int = None) -> Coroutine:
        """Subscribes on market data of specified symbol (see
        https://metaapi.cloud/docs/client/websocket/marketDataStreaming/subscribeToMarketData/).

        Args:
            symbol: Symbol (e.g. currency pair or an index).
            instance_index: Instance index.

        Returns:
            Promise which resolves when subscription request was processed.
        """
        self._subscriptions[symbol] = True
        return self._websocketClient.subscribe_to_market_data(self._account.id, instance_index, symbol)

    def unsubscribe_from_market_data(self, symbol: str, instance_index: int = None) -> Coroutine:
        """Unsubscribes from market data of specified symbol (see
        https://metaapi.cloud/docs/client/websocket/marketDataStreaming/subscribeToMarketData/).

        Args:
            symbol: Symbol (e.g. currency pair or an index).
            instance_index: Instance index.

        Returns:
            Promise which resolves when subscription request was processed.
        """
        if symbol in self._subscriptions:
            del self._subscriptions[symbol]
        return self._websocketClient.unsubscribe_from_market_data(self._account.id, instance_index, symbol)

    @property
    def subscribed_symbols(self) -> List[str]:
        """Returns list of the symbols connection is subscribed to.

        Returns:
            List of the symbols connection is subscribed to.
        """
        return list(self._subscriptions.keys())

    def get_symbol_specification(self, symbol: str) -> 'Coroutine[asyncio.Future[MetatraderSymbolSpecification]]':
        """Retrieves specification for a symbol (see
        https://metaapi.cloud/docs/client/websocket/api/retrieveMarketData/getSymbolSpecification/).

        Args:
            symbol: Symbol to retrieve specification for.

        Returns:
            A coroutine which resolves when specification MetatraderSymbolSpecification is retrieved.
        """
        return self._websocketClient.get_symbol_specification(self._account.id, symbol)

    def get_symbol_price(self, symbol) -> 'Coroutine[asyncio.Future[MetatraderSymbolPrice]]':
        """Retrieves specification for a symbol (see
        https://metaapi.cloud/docs/client/websocket/api/retrieveMarketData/getSymbolPrice/).

        Args:
            symbol: Symbol to retrieve price for.

        Returns:
            A coroutine which resolves when price MetatraderSymbolPrice is retrieved.
        """
        return self._websocketClient.get_symbol_price(self._account.id, symbol)

    def save_uptime(self, uptime: Dict):
        """Sends client uptime stats to the server.

        Args:
            uptime: Uptime statistics to send to the server.

        Returns:
            A coroutine which resolves when uptime statistics is submitted.
        """
        return self._websocketClient.save_uptime(self._account.id, uptime)

    @property
    def terminal_state(self) -> TerminalState:
        """Returns local copy of terminal state.

        Returns:
            Local copy of terminal state.
        """
        return self._terminalState

    @property
    def history_storage(self) -> HistoryStorage:
        """Returns local history storage.

        Returns:
            Local history storage.
        """
        return self._historyStorage

    def add_synchronization_listener(self, listener):
        """Adds synchronization listener.

        Args:
            listener: Synchronization listener to add.
        """
        self._websocketClient.add_synchronization_listener(self._account.id, listener)

    def remove_synchronization_listener(self, listener):
        """Removes synchronization listener for specific account.

        Args:
            listener: Synchronization listener to remove.
        """
        self._websocketClient.remove_synchronization_listener(self._account.id, listener)

    async def on_connected(self, instance_index: int, replicas: int):
        """Invoked when connection to MetaTrader terminal established.

        Args:
            instance_index: Index of an account instance connected.
            replicas: Number of account replicas launched.

        Returns:
            A coroutine which resolves when the asynchronous event is processed.
        """
        if self._subscribeFuture and not self._subscribeFuture.done():
            self._subscribeFuture.set_result(False)
            self._subscribeTask.cancel()
        self._shouldRetrySubscribe = False
        key = random_id(32)
        state = self._get_state(instance_index)
        state['shouldSynchronize'] = key
        state['synchronizationRetryIntervalInSeconds'] = 1
        state['synchronized'] = False
        await self._ensure_synchronized(instance_index, key)
        indices = []
        for i in range(replicas):
            indices.append(i)
        for key in list(self._stateByInstanceIndex.keys()):
            e = self._stateByInstanceIndex[key]
            if e['instanceIndex'] not in indices:
                del self._stateByInstanceIndex[key]

    async def on_disconnected(self, instance_index: int):
        """Invoked when connection to MetaTrader terminal terminated.

        Args:
            instance_index: Index of an account instance connected.

        Returns:
             A coroutine which resolves when the asynchronous event is processed.
        """
        state = self._get_state(instance_index)
        state['lastDisconnectedSynchronizationId'] = state['lastSynchronizationId']
        state['lastSynchronizationId'] = None
        state['shouldSynchronize'] = None
        state['synchronized'] = False
        state['disconnected'] = True

    async def on_deal_synchronization_finished(self, instance_index: int, synchronization_id: str):
        """Invoked when a synchronization of history deals on a MetaTrader account have finished.

        Args:
            instance_index: Index of an account instance connected.
            synchronization_id: Synchronization request id.

        Returns:
            A coroutine which resolves when the asynchronous event is processed.
        """
        state = self._get_state(instance_index)
        state['dealsSynchronized'][synchronization_id] = True

    async def on_order_synchronization_finished(self, instance_index: int, synchronization_id: str):
        """Invoked when a synchronization of history orders on a MetaTrader account have finished.

        Args:
            instance_index: Index of an account instance connected.
            synchronization_id: Synchronization request id.

        Returns:
             A coroutine which resolves when the asynchronous event is processed
        """
        state = self._get_state(instance_index)
        state['ordersSynchronized'][synchronization_id] = True

    async def on_reconnected(self):
        """Invoked when connection to MetaApi websocket API restored after a disconnect.

        Returns:
            A coroutine which resolves when connection to MetaApi websocket API restored after a disconnect.
        """
        if self._subscribeFuture and not self._subscribeFuture.done():
            self._subscribeFuture.set_result(False)
            self._subscribeTask.cancel()
        await self.subscribe()

    async def is_synchronized(self, instance_index: int, synchronization_id: str = None) -> bool:
        """Returns flag indicating status of state synchronization with MetaTrader terminal.

        Args:
            instance_index: Index of an account instance connected.
            synchronization_id: Optional synchronization request id, last synchronization request id will be used.

        Returns:
            A coroutine resolving with a flag indicating status of state synchronization with MetaTrader terminal.
        """
        def reducer_func(acc, s: MetaApiConnectionDict):
            if instance_index is not None and s['instanceIndex'] != instance_index:
                return acc
            nonlocal synchronization_id
            synchronization_id = synchronization_id or s['lastSynchronizationId']
            synchronized = synchronization_id in s['ordersSynchronized'] and \
                bool(s['ordersSynchronized'][synchronization_id]) and \
                synchronization_id in s['dealsSynchronized'] and \
                bool(s['dealsSynchronized'][synchronization_id])
            return acc or synchronized

        return reduce(reducer_func, self._stateByInstanceIndex.values(), False) if \
            len(self._stateByInstanceIndex.values()) else False

    async def wait_synchronized(self, opts: SynchronizationOptions = None):
        """Waits until synchronization to MetaTrader terminal is completed.

        Args:
            opts: Synchronization options.

        Returns:
            A coroutine which resolves when synchronization to MetaTrader terminal is completed.

        Raises:
            TimeoutException: If application failed to synchronize with the terminal within timeout allowed.
        """
        start_time = datetime.now()
        opts = opts or {}
        instance_index = opts['instanceIndex'] if 'instanceIndex' in opts else None
        synchronization_id = opts['synchronizationId'] if 'synchronizationId' in opts else None
        timeout_in_seconds = opts['timeoutInSeconds'] if 'timeoutInSeconds' in opts else 300
        interval_in_milliseconds = opts['intervalInMilliseconds'] if 'intervalInMilliseconds' in opts else 1000
        application_pattern = opts['applicationPattern'] if 'applicationPattern' in opts \
            else ('CopyFactory.*|RPC' if self._account.application == 'CopyFactory' else 'RPC')
        synchronized = await self.is_synchronized(instance_index, synchronization_id)
        while not synchronized and (start_time + timedelta(seconds=timeout_in_seconds) > datetime.now()):
            await asyncio.sleep(interval_in_milliseconds / 1000)
            synchronized = await self.is_synchronized(instance_index, synchronization_id)
        state = None
        if instance_index is None:
            for s in self._stateByInstanceIndex.values():
                if await self.is_synchronized(s['instanceIndex'], synchronization_id):
                    state = s
                    instance_index = s['instanceIndex']
        else:
            state = next((s for s in self._stateByInstanceIndex if s['instanceIndex'] == instance_index), None)
        if not synchronized:
            raise TimeoutException('Timed out waiting for MetaApi to synchronize to MetaTrader account ' +
                                   self._account.id + ', synchronization id ' +
                                   (synchronization_id or (bool(state) and state['lastSynchronizationId']) or
                                    (bool(state) and state['lastDisconnectedSynchronizationId']) or 'None'))
        time_left_in_seconds = max(0, timeout_in_seconds - (datetime.now() - start_time).total_seconds())
        await self._websocketClient.wait_synchronized(self._account.id, instance_index, application_pattern,
                                                      time_left_in_seconds)

    async def close(self):
        """Closes the connection. The instance of the class should no longer be used after this method is invoked."""
        if not self._closed:
            self._stateByInstanceIndex = {}
            await self._websocketClient.unsubscribe(self._account.id)
            self._websocketClient.remove_synchronization_listener(self._account.id, self)
            self._websocketClient.remove_synchronization_listener(self._account.id, self._terminalState)
            self._websocketClient.remove_synchronization_listener(self._account.id, self._historyStorage)
            self._websocketClient.remove_synchronization_listener(self._account.id, self._healthMonitor)
            self._connection_registry.remove(self._account.id)
            self._healthMonitor.stop()
            self._closed = True

    @property
    def synchronized(self) -> bool:
        """Returns synchronization status.

        Returns:
            Synchronization status.
        """
        return True in list(map(lambda s: s['synchronized'], self._stateByInstanceIndex.values()))

    @property
    def account(self) -> MetatraderAccountModel:
        """Returns MetaApi account.

        Returns:
            MetaApi account.
        """
        return self._account

    @property
    def health_monitor(self) -> ConnectionHealthMonitor:
        """Returns connection health monitor instance.

        Returns:
            Connection health monitor instance.
        """
        return self._healthMonitor

    async def _ensure_synchronized(self, instance_index: int, key):
        state = self._get_state(instance_index)
        if state:
            try:
                await self.synchronize(instance_index)
                for symbol in self.subscribed_symbols:
                    await self.subscribe_to_market_data(symbol, instance_index)
                state['synchronized'] = True
                state['synchronizationRetryIntervalInSeconds'] = 1
            except Exception as err:
                print(f'[{datetime.now().isoformat()}] MetaApi websocket client for account ' + self.account.id +
                      ':' + str(instance_index) + ' failed to synchronize', err)
                if state['shouldSynchronize'] == key:

                    async def restart_ensure_sync():
                        await asyncio.sleep(state['synchronizationRetryIntervalInSeconds'])
                        await self._ensure_synchronized(instance_index, key)
                    asyncio.create_task(restart_ensure_sync())
                    state['synchronizationRetryIntervalInSeconds'] = \
                        min(state['synchronizationRetryIntervalInSeconds'] * 2, 300)

    def _get_state(self, instance_index: int) -> MetaApiConnectionDict:
        if str(instance_index) not in self._stateByInstanceIndex:
            self._stateByInstanceIndex[str(instance_index)] = {
                'instanceIndex': instance_index,
                'ordersSynchronized': {},
                'dealsSynchronized': {},
                'shouldSynchronize': None,
                'synchronizationRetryIntervalInSeconds': 1,
                'synchronized': False,
                'lastDisconnectedSynchronizationId': None,
                'lastSynchronizationId': None,
                'disconnected': False
            }
        return self._stateByInstanceIndex[str(instance_index)]



from .. import MetaApi
import os
import pytest
import asyncio
import json

token = os.getenv('TOKEN')
login = os.getenv('LOGIN')
password = os.getenv('PASSWORD')
server_name = os.getenv('SERVER')
server_dat_file = os.getenv('PATH_TO_SERVERS_DAT')


class TestDoubleSync:
    @pytest.mark.asyncio
    async def test_not_corrupt(self):
        """Should not corrupt files after simultaneous synchronization."""
        if token:
            if not os.path.exists('.metaapi'):
                os.mkdir('.metaapi')
            api = MetaApi(token, {'application': 'MetaApi', 'domain': 'project-stock.v2.agiliumlabs.cloud'})
            profiles = await api.provisioning_profile_api.get_provisioning_profiles()
            profile = None
            for item in profiles:
                if item.name == server_name:
                    profile = item
                    break
            if not profile:
                profile = await api.provisioning_profile_api.create_provisioning_profile({
                    'name': server_name,
                    'version': 5,
                    'brokerTimezone': 'EET',
                    'brokerDSTSwitchTimezone': 'EET'
                })
                await profile.upload_file('servers.dat', server_dat_file)
            if profile and profile.status == 'new':
                await profile.upload_file('servers.dat', server_dat_file)
            accounts = await api.metatrader_account_api.get_accounts()
            account = None
            for item in accounts:
                if item.login == login and item.type.startswith('cloud'):
                    account = item
                    break
            if not account:
                account = await api.metatrader_account_api.create_account({
                    'name': 'Test account',
                    'type': 'cloud',
                    'login': login,
                    'password': password,
                    'server': server_name,
                    'provisioningProfileId': profile.id,
                    'application': 'MetaApi',
                    'magic': 1000
                })
            account_copy = await api.metatrader_account_api.get_account(account.id)
            await asyncio.gather(*[
                account.deploy(),
                account_copy.deploy()
            ])
            await asyncio.gather(*[
                account.wait_connected(),
                account_copy.wait_connected()
            ])
            connection = await account.connect()
            connection_copy = await account_copy.connect()
            await asyncio.gather(*[
                connection.wait_synchronized({'timeoutInSeconds': 1200}),
                connection_copy.wait_synchronized({'timeoutInSeconds': 1200})
            ])
            await account.undeploy()
            await account_copy.undeploy()
            api._metaApiWebsocketClient.remove_all_listeners()
            json.loads(open(f'.metaapi/{account.id}-{account.application}-deals.bin').read())
            json.loads(open(f'.metaapi/{account.id}-{account.application}-historyOrders.bin').read())



from ..metaApi_client import MetaApiClient
from typing_extensions import TypedDict
from typing import List, Union, Optional, Dict
from httpx import Response
from enum import Enum


class Extension(TypedDict):
    """Extension model."""
    id: str
    """Extension id."""
    configuration: Dict
    """Extension configuration."""


class State(Enum):
    """Account state."""
    CREATED = 'CREATED'
    DEPLOYING = 'DEPLOYING'
    DEPLOYED = 'DEPLOYED'
    DEPLOY_FAILED = 'DEPLOY_FAILED'
    UNDEPLOYING = 'UNDEPLOYING'
    UNDEPLOYED = 'UNDEPLOYED'
    UNDEPLOY_FAILED = 'UNDEPLOY_FAILED'
    DELETING = 'DELETING'
    DELETE_FAILED = 'DELETE_FAILED'
    REDEPLOY_FAILED = 'REDEPLOY_FAILED'


class ConnectionStatus(Enum):
    """Account connection status."""
    CONNECTED = 'CONNECTED'
    DISCONNECTED = 'DISCONNECTED'
    DISCONNECTED_FROM_BROKER = 'DISCONNECTED_FROM_BROKER'


class AccountsFilter(TypedDict):

    offset: Optional[int]
    """Search offset (defaults to 0) (must be greater or equal to 0)."""
    limit: Optional[int]
    """Search limit (defaults to 1000) (must be greater or equal to 1 and less or equal to 1000)."""
    version: Optional[Union[List[int], int]]
    """MT version (allowed values are 4 and 5)"""
    type: Optional[Union[List[str], str]]
    """Account type. Allowed values are 'cloud' and 'self-hosted'"""
    state: Optional[Union[List[State], State]]
    """Account state."""
    connectionStatus: Optional[Union[List[ConnectionStatus], ConnectionStatus]]
    """Connection status."""
    query: Optional[str]
    """Searches over _id, name, server and login to match query."""
    provisioningProfileId: Optional[str]
    """Provisioning profile id."""


class MetatraderAccountIdDto(TypedDict):
    """MetaTrader account id model"""

    id: str
    """MetaTrader account unique identifier"""


class MetatraderAccountDto(TypedDict):
    """MetaTrader account model"""

    _id: str
    """Account unique identifier."""
    name: str
    """MetaTrader account human-readable name in the MetaApi app."""
    type: str
    """Account type, can be cloud, cloud-g1, cloud-g2 or self-hosted. Cloud and cloud-g2 are aliases."""
    login: str
    """MetaTrader account number."""
    server: str
    """MetaTrader server which hosts the account."""
    provisioningProfileId: str
    """Id of the account's provisioning profile."""
    application: str
    """Application name to connect the account to. Currently allowed values are MetaApi and AgiliumTrade"""
    magic: int
    """MetaTrader magic to place trades using."""
    state: str
    """Account deployment state. One of CREATED, DEPLOYING, DEPLOYED, UNDEPLOYING, UNDEPLOYED, DELETING"""
    connectionStatus: str
    """Terminal & broker connection status, one of CONNECTED, DISCONNECTED, DISCONNECTED_FROM_BROKER"""
    accessToken: str
    """Authorization token to be used for accessing single account data. Intended to be used in browser API."""
    manualTrades: bool
    """Flag indicating if trades should be placed as manual trades. Default is false."""
    quoteStreamingIntervalInSeconds: float
    """Quote streaming interval in seconds. Set to 0 in order to receive quotes on each tick. Default value is
    2.5 seconds. Intervals less than 2.5 seconds are supported only for G2."""
    tags: Optional[List[str]]
    """MetaTrader account tags."""
    extensions: List[Extension]
    """API extensions."""
    metadata: Dict
    """Extra information which can be stored together with your account."""
    reliability: str
    """Used to increase the reliability of the account. Allowed values are regular and high. Default is regular."""


class NewMetatraderAccountDto(TypedDict):
    """New MetaTrader account model"""

    name: str
    """MetaTrader account human-readable name in the MetaApi app."""
    type: str
    """Account type, can be cloud, cloud-g1, cloud-g2 or self-hosted. cloud-g2 and cloud are aliases. When you
    create MT5 cloud account the type is automatically converted to cloud-g1 because MT5 G2 support is still
    experimental. You can still create MT5 G2 account by setting type to cloud-g2."""
    login: str
    """MetaTrader account number."""
    password: str
    """MetaTrader account password. The password can be either investor password for read-only access or master
    password to enable trading features. Required for cloud account."""
    server: str
    """MetaTrader server which hosts the account."""
    provisioningProfileId: str
    """Id of the account's provisioning profile."""
    application: str
    """Application name to connect the account to. Currently allowed values are MetaApi and AgiliumTrade."""
    magic: int
    """MetaTrader magic to place trades using."""
    manualTrades: bool
    """Flag indicating if trades should be placed as manual trades. Default is false."""
    quoteStreamingIntervalInSeconds: float
    """Quote streaming interval in seconds. Set to 0 in order to receive quotes on each tick. Default value is
    2.5 seconds. Intervals less than 2.5 seconds are supported only for G2."""
    tags: Optional[List[str]]
    """MetaTrader account tags."""
    extensions: List[Extension]
    """API extensions."""
    metadata: Dict
    """Extra information which can be stored together with your account."""
    reliability: str
    """Used to increase the reliability of the account. Allowed values are regular and high. Default is regular."""


class MetatraderAccountUpdateDto(TypedDict):
    """Updated MetaTrader account data"""

    name: str
    """MetaTrader account human-readable name in the MetaApi app."""
    password: str
    """MetaTrader account password. The password can be either investor password for read-only
    access or master password to enable trading features. Required for cloud account"""
    server: str
    """MetaTrader server which hosts the account"""
    manualTrades: bool
    """Flag indicating if trades should be placed as manual trades. Default is false."""
    quoteStreamingIntervalInSeconds: float
    """Quote streaming interval in seconds. Set to 0 in order to receive quotes on each tick. Default value is
    2.5 seconds. Intervals less than 2.5 seconds are supported only for G2."""
    tags: Optional[List[str]]
    """MetaTrader account tags."""
    extensions: List[Extension]
    """API extensions."""
    metadata: Dict
    """Extra information which can be stored together with your account."""


class MetatraderAccountClient(MetaApiClient):
    """metaapi.cloud MetaTrader account API client (see https://metaapi.cloud/docs/provisioning/)

    Attributes:
        _httpClient: HTTP client
        _host: domain to connect to
        _token: authorization token
    """

    async def get_accounts(self, accounts_filter: AccountsFilter = None) -> Response:
        """Retrieves MetaTrader accounts owned by user
        (see https://metaapi.cloud/docs/provisioning/api/account/readAccounts/)

        Args:
            accounts_filter: Optional filter.

        Returns:
            A coroutine resolving with List[MetatraderAccountDto] - MetaTrader accounts found.
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('get_accounts')
        opts = {
            'url': f'{self._host}/users/current/accounts',
            'method': 'GET',
            'params': accounts_filter or {},
            'headers': {
                'auth-token': self._token
            }
        }
        return await self._httpClient.request(opts)

    async def get_account(self, id: str) -> Response:
        """Retrieves a MetaTrader account by id (see https://metaapi.cloud/docs/provisioning/api/account/readAccount/).
        Throws an error if account is not found.

        Args:
            id: MetaTrader account id.

        Returns:
            A coroutine resolving with MetatraderAccountDto - MetaTrader account found.
        """
        opts = {
            'url': f'{self._host}/users/current/accounts/{id}',
            'method': 'GET',
            'headers': {
                'auth-token': self._token
            }
        }
        return await self._httpClient.request(opts)

    async def get_account_by_token(self) -> 'Response[MetatraderAccountDto]':
        """Retrieves a MetaTrader account by token
        (see https://metaapi.cloud/docs/provisioning/api/account/readAccount/). Throws an error if account is
        not found. Method is accessible only with account access token.

        Returns:
            A coroutine resolving with MetaTrader account found.
        """
        if self._is_not_account_token():
            return self._handle_no_access_exception('get_account_by_token')
        opts = {
            'url': f'{self._host}/users/current/accounts/accessToken/{self._token}',
            'method': 'GET'
        }
        return await self._httpClient.request(opts)

    async def create_account(self, account: NewMetatraderAccountDto) -> Response:
        """Starts cloud API server for a MetaTrader account using specified provisioning profile
        (see https://metaapi.cloud/docs/provisioning/api/account/createAccount/).
        It takes some time to launch the terminal and connect the terminal to the broker, you can use the
        connectionStatus field to monitor the current status of the terminal.

        Args:
            account: MetaTrader account to create.

        Returns:
            A coroutine resolving with MetatraderAccountIdDto - an id of the MetaTrader account created.
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('create_account')
        opts = {
            'url': f'{self._host}/users/current/accounts',
            'method': 'POST',
            'headers': {
                'auth-token': self._token
            },
            'body': account
        }
        return await self._httpClient.request(opts)

    async def deploy_account(self, id: str) -> Response:
        """Starts API server for MetaTrader account. This request will be ignored if the account has already
        been deployed. (see https://metaapi.cloud/docs/provisioning/api/account/deployAccount/)

        Args:
            id: MetaTrader account id to deploy.

        Returns:
            A coroutine resolving when MetaTrader account is scheduled for deployment
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('deploy_account')
        opts = {
            'url': f'{self._host}/users/current/accounts/{id}/deploy',
            'method': 'POST',
            'headers': {
                'auth-token': self._token
            }
        }
        return await self._httpClient.request(opts)

    async def undeploy_account(self, id: str) -> Response:
        """Stops API server for a MetaTrader account. Terminal data such as downloaded market history data will
        be preserved. (see https://metaapi.cloud/docs/provisioning/api/account/undeployAccount/)

        Args:
            id: MetaTrader account id to undeploy.

        Returns:
            A coroutine resolving when MetaTrader account is scheduled for undeployment.
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('undeploy_account')
        opts = {
            'url': f'{self._host}/users/current/accounts/{id}/undeploy',
            'method': 'POST',
            'headers': {
                'auth-token': self._token
            }
        }
        return await self._httpClient.request(opts)

    async def redeploy_account(self, id: str) -> Response:
        """Redeploys MetaTrader account. This is equivalent to undeploy immediately followed by deploy.
        (see https://metaapi.cloud/docs/provisioning/api/account/redeployAccount/)

        Args:
            id: MetaTrader account id to redeploy.

        Returns:
            A coroutine resolving when MetaTrader account is scheduled for redeployment.
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('redeploy_account')
        opts = {
            'url': f'{self._host}/users/current/accounts/{id}/redeploy',
            'method': 'POST',
            'headers': {
                'auth-token': self._token
            }
        }
        return await self._httpClient.request(opts)

    async def delete_account(self, id: str) -> Response:
        """Stops and deletes an API server for a specified MetaTrader account. The terminal state such as downloaded
        market data history will be deleted as well when you delete the account.
        (see https://metaapi.cloud/docs/provisioning/api/account/deleteAccount/)

        Args:
            id: MetaTrader account id to delete.

        Returns:
            A coroutine resolving when MetaTrader account is scheduled for deletion.
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('delete_account')
        opts = {
            'url': f'{self._host}/users/current/accounts/{id}',
            'method': 'DELETE',
            'headers': {
                'auth-token': self._token
            }
        }
        return await self._httpClient.request(opts)

    async def update_account(self, id: str, account: MetatraderAccountUpdateDto) -> Response:
        """Updates existing metatrader account data (see
        https://metaapi.cloud/docs/provisioning/api/account/updateAccount/)

        Args:
            id: MetaTrader account id.
            account: Updated MetaTrader account.

        Returns:
            A coroutine resolving when MetaTrader account is updated.
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('update_account')
        opts = {
            'url': f'{self._host}/users/current/accounts/{id}',
            'method': 'PUT',
            'headers': {
                'auth-token': self._token
            },
            'body': account
        }
        return await self._httpClient.request(opts)
